from __future__ import division, print_function
import matplotlib.pyplot as plt
import numpy as np
import scipy.linalg
np.set_printoptions(precision=4)
from control import *

#dlqr idea based on here: http://www.mwm.im/lqr-controllers-with-python/
def dlqr(A,B,Q,R):
    #ref Bertsekas, p.151
    #X = np.matrix(scipy.linalg.solve_discrete_are(A, B, Q, R))
    X,L,E = dare(A,B,Q,R)
    #compute the LQR gain
    K = np.matrix(scipy.linalg.inv(B.T*X*B+R)*(B.T*X*A))
    eigVals, eigVecs = scipy.linalg.eig(A-B*K)
    return K, X, eigVals

K_e = 0.0053 #from data
K_m =  K_e#motor coefficient
K_t = 1.78e-3 #from data
T = 0.001 #in seconds
J_m = 5e-6 #N*m
J_a = 4.5e-4  #N*m
R_m = 0.5 #in ohms
R_s = 0.5 #in ohms
L_a = 0.15 #in meters (15 cm)
K_d =0 #1e-2 drag on prop from arm movement...consider changing if you want!
K_f = 0 #friction coefficient consider changing if you want!


#three state plant model third state is Ke*omega
A = np.matrix([[1,T,0],[0,1,T*K_t*L_a/(J_a*K_e)],[0,K_d,1-T*K_m*K_e/(J_m*(R_m+R_s))-T*K_f]])
B = np.matrix([[0],[0],[T*K_e*K_m/(J_m*(R_s+R_m))]])
C = np.matrix([1,0,0])
D = np.matrix([0])
sys3 = ss(A,B,C,D,T)

#create four state model (with integrator) from three state:
A_top = np.concatenate((A,np.zeros(shape=(3,1))), axis=1)
A_bot = np.concatenate((T*C,np.matrix([[1]])), axis=1)
A_til = np.concatenate((A_top,A_bot),axis=0)
#start off B with 0 in 1,4 spot (so isn't properly B tilda):
B_til = np.concatenate((B,np.matrix([[0]])),axis=0)
C_til = np.concatenate((C,np.matrix([[0]])),axis=1)
D_til = D

sys4 = ss(A_til,B_til,C_til,D_til,T)


#acker natural frequency picking for three states:
des_nat_freq_3 = [0.99,0.99,0.99]
k_acker_3 = acker(A,B,des_nat_freq_3)
print(k_acker_3)

#acker natural frequency picking for four states:
des_nat_freq_4 = [0.95,0.95,0.95,0.95]
k_acker_4 = acker(A_til,B_til,des_nat_freq_4)
print(k_acker_4)

k_3 = k_acker_3 #change to k_acker or manually set as needed
k_4 = k_acker_4


#calculate K_u for three state system:
KuInv = C*np.linalg.inv(np.eye(3)-1*(A-B*k_3))*B
Ku3 = np.linalg.inv(KuInv)
print(Ku3)

#calculate K_u for four state system:
KuInv = C*np.linalg.inv(np.eye(3)-1*(A-B*np.matrix(k_4.A[0][0:3])))*B
Ku4 = np.linalg.inv(KuInv)
print(Ku4)


#closed loop three-state equivalent:
Ac3 = A-B*k_3 #equivalent to M in notes
Bc3 = B*Ku3
Cc3 = C
Dc3 = D
overall3=ss(Ac3,Bc3,Cc3,Dc3,T)


#closed loop four state equivalents:
Ac4 = A_til - B_til*k_4 #equivalent to M
B_til[(3,0)] = -T/Ku4
Bc4 = B_til*Ku4
Cc4 = C_til
Dc4 = D_til
overall4 = ss(Ac4,Bc4,Cc4,Dc4,T)



#simulate four state system:
t = np.arange(0,3,0.1)
sig = 0.1*np.ones(np.size(t))
y,t1,x=lsim(overall4,sig,t)

plot_in_radians = False
if plot_in_radians:
    scale = 1
else:
    scale = 180/3.14159

command = np.array(Ku4*sig[1] - k_4*x)
plt.subplot(5, 1, 1)
x1, = plt.plot(scale*x[0], label='theta_a')
plt.ylabel('theta_a')
plt.subplot(5, 1, 2)
x2, = plt.plot(scale*x[1], label='omega_a')
plt.ylabel('omega_a')
plt.subplot(5, 1, 3)
x3, = plt.plot(x[2], label='K_e*omega_m')
plt.ylabel('back_emf')
plt.subplot(5, 1, 4)
x3, = plt.plot(scale*x[3], label='integral')
plt.ylabel('integral')
plt.subplot(5, 1, 5)
com, = plt.plot(command[0], label='command')
plt.ylabel('command')
plt.xlabel('time steps')
plt.show()

